import{S as A,i as B,s as E,D as d,Z as o,_ as g,m,h as f,n as R,$ as _,b as w,a0 as z,K as k,a1 as c,M as v,Q as b,q as Y,r as q,u as D}from"./index.ad1b71d6.js";function M(a){let e,l;return{c(){e=o("title"),l=Y(a[1])},l(t){e=g(t,"title",{});var r=m(e);l=q(r,a[1]),r.forEach(f)},m(t,r){w(t,e,r),z(e,l)},p(t,r){r&2&&D(l,t[1])},d(t){t&&f(e)}}}function L(a){let e,l,t=a[1]&&M(a),r=[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},{width:a[0]},{height:a[0]},a[2],a[3]],n={};for(let i=0;i<r.length;i+=1)n=d(n,r[i]);return{c(){e=o("svg"),t&&t.c(),l=o("path"),this.h()},l(i){e=g(i,"svg",{xmlns:!0,viewBox:!0,fill:!0,preserveAspectRatio:!0,width:!0,height:!0});var s=m(e);t&&t.l(s),l=g(s,"path",{d:!0}),m(l).forEach(f),s.forEach(f),this.h()},h(){R(l,"d","M16 22L6 12 7.4 10.6 16 19.2 24.6 10.6 26 12z"),_(e,n)},m(i,s){w(i,e,s),t&&t.m(e,null),z(e,l)},p(i,[s]){i[1]?t?t.p(i,s):(t=M(i),t.c(),t.m(e,l)):t&&(t.d(1),t=null),_(e,n=k(r,[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},s&1&&{width:i[0]},s&1&&{height:i[0]},s&4&&i[2],s&8&&i[3]]))},i:c,o:c,d(i){i&&f(e),t&&t.d()}}}function N(a,e,l){let t,r;const n=["size","title"];let i=v(e,n),{size:s=16}=e,{title:h=void 0}=e;return a.$$set=u=>{l(5,e=d(d({},e),b(u))),l(3,i=v(e,n)),"size"in u&&l(0,s=u.size),"title"in u&&l(1,h=u.title)},a.$$.update=()=>{l(4,t=e["aria-label"]||e["aria-labelledby"]||h),l(2,r={"aria-hidden":t?void 0:!0,role:t?"img":void 0,focusable:Number(e.tabindex)===0?!0:void 0})},e=b(e),[s,h,r,i,t]}class Q extends A{constructor(e){super(),B(this,e,N,L,E,{size:0,title:1})}}function C(a){let e,l;return{c(){e=o("title"),l=Y(a[1])},l(t){e=g(t,"title",{});var r=m(e);l=q(r,a[1]),r.forEach(f)},m(t,r){w(t,e,r),z(e,l)},p(t,r){r&2&&D(l,t[1])},d(t){t&&f(e)}}}function P(a){let e,l,t=a[1]&&C(a),r=[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},{width:a[0]},{height:a[0]},a[2],a[3]],n={};for(let i=0;i<r.length;i+=1)n=d(n,r[i]);return{c(){e=o("svg"),t&&t.c(),l=o("path"),this.h()},l(i){e=g(i,"svg",{xmlns:!0,viewBox:!0,fill:!0,preserveAspectRatio:!0,width:!0,height:!0});var s=m(e);t&&t.l(s),l=g(s,"path",{d:!0}),m(l).forEach(f),s.forEach(f),this.h()},h(){R(l,"d","M16 10L26 20 24.6 21.4 16 12.8 7.4 21.4 6 20z"),_(e,n)},m(i,s){w(i,e,s),t&&t.m(e,null),z(e,l)},p(i,[s]){i[1]?t?t.p(i,s):(t=C(i),t.c(),t.m(e,l)):t&&(t.d(1),t=null),_(e,n=k(r,[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},s&1&&{width:i[0]},s&1&&{height:i[0]},s&4&&i[2],s&8&&i[3]]))},i:c,o:c,d(i){i&&f(e),t&&t.d()}}}function S(a,e,l){let t,r;const n=["size","title"];let i=v(e,n),{size:s=16}=e,{title:h=void 0}=e;return a.$$set=u=>{l(5,e=d(d({},e),b(u))),l(3,i=v(e,n)),"size"in u&&l(0,s=u.size),"title"in u&&l(1,h=u.title)},a.$$.update=()=>{l(4,t=e["aria-label"]||e["aria-labelledby"]||h),l(2,r={"aria-hidden":t?void 0:!0,role:t?"img":void 0,focusable:Number(e.tabindex)===0?!0:void 0})},e=b(e),[s,h,r,i,t]}class U extends A{constructor(e){super(),B(this,e,S,P,E,{size:0,title:1})}}export{Q as C,U as a};
