.removable-tag{max-width:12rem}.filter-pill .bx--tooltip__label{margin-right:.25rem;display:inline-block;height:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap;max-width:5rem}.filter-pill .bx--tooltip__content{display:flex;flex-direction:column;align-items:center}.active.svelte-1ovvpwo{border-radius:.5rem;--tw-bg-opacity:1;background-color:rgb(245 245 245 / var(--tw-bg-opacity));outline-style:solid;outline-width:1px;outline-color:#a3a3a3}.drop-pill .bx--list-box__menu{max-height:26rem!important;width:unset}.drop-pill.drop-pill-left .bx--list-box__menu{transform:translate(-100%);margin-left:100%}.drop-pill bx--list-box--expanded{display:flex;flex-direction:column}.drop-pill .bx--dropdown{max-width:20rem!important}.drop-pill .bx--dropdown__wrapper--inline{gap:0px!important}.drop-pill .bx--list-box__menu-item__option{padding-right:0}.drop-pill .bx--list-box__menu-icon{position:unset}.drop-pill .bx--list-box__field{padding-right:0!important}.search-pill .bx--tooltip__label{margin-right:.25rem;display:inline-block;height:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap;max-width:5rem}.search-pill .bx--tooltip__content{display:flex;flex-direction:column;align-items:center}.filter-group.svelte-jhakxs{min-width:6rem;display:flex;flex-direction:row;align-items:center;-moz-column-gap:.5rem;column-gap:.5rem;padding-left:.5rem;padding-right:.5rem;--tw-shadow:0 1px 2px 0 rgb(0 0 0 / .05);--tw-shadow-colored:0 1px 2px 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.search-icon.svelte-m5d7d0{background-color:#fafafa}.bx--form__helper-text{padding:0 0 0 1rem}.compute-embedding .bx--btn{height:3rem}.compute-embedding-indexing .bx--btn.bx--btn--disabled{color:transparent}.embedding-select .bx--select,.field-select .bx--select{flex-direction:row}.new-concept .bx--tag,.new-keyword .bx--tag{width:1.5rem;min-width:0px;padding-left:0;padding-right:0}.new-concept,.new-keyword{height:100%}.bx--list-box__menu-item:not(.bx--list-box__menu-item--highlighted):has(.isSignal){--tw-bg-opacity:1;background-color:rgb(239 246 255 / var(--tw-bg-opacity))}.bx--list-box__menu-item:not(.bx--list-box__menu-item--highlighted):has(.isConcept){--tw-bg-opacity:1;background-color:rgb(209 250 229 / var(--tw-bg-opacity))}.bx--list-box__menu-item:not(.bx--list-box__menu-item--highlighted):has(.isLabel){--tw-bg-opacity:1;background-color:rgb(204 251 241 / var(--tw-bg-opacity))}.bx--list-box__menu-item:not(.bx--list-box__menu-item--highlighted):has(.isExcludeLabel){--tw-bg-opacity:1;background-color:rgb(254 226 226 / var(--tw-bg-opacity))}.search-container .bx--list-box__menu{max-height:26rem!important}.search-container .bx--list-box__field{height:100%}.embedding-dropdown .bx--list-box__menu-item__option{padding-right:0}ul.bx--overflow-menu-options{width:11rem}.signal-badge .bx--tag{padding-left:.5rem;padding-right:.5rem}.signal-badge .bx--tag__custom-icon{margin-right:0}.signal-badge .bx--tooltip__label .bx--tooltip__trigger{margin:0}.histogram-label.svelte-1yohsr2{font-family:inconsolata variable}.stats-table.svelte-1k80rmc td.svelte-1k80rmc{width:12rem}.stats-table.svelte-1k80rmc td.svelte-1k80rmc:first-child{overflow:hidden;text-overflow:ellipsis;white-space:nowrap;padding-top:.5rem;padding-bottom:.5rem;padding-right:.5rem}.stats-table.svelte-1k80rmc td.svelte-1k80rmc:last-child{overflow:hidden;text-overflow:ellipsis;white-space:nowrap;padding-top:.5rem;padding-bottom:.5rem;padding-left:.5rem}.stats-label.svelte-1k80rmc.svelte-1k80rmc{font-family:inconsolata variable}.signal-tag span{padding-left:.5rem;padding-right:.5rem}.compute-signal-chip .bx--tooltip__label .bx--tooltip__trigger{margin:0}.compute-signal-preview .bx--tag{cursor:pointer}.compute-signal-preview .bx--tag__custom-icon{margin:0}.schema .bx--tab-content{padding:0!important}.card-title.svelte-1tyx0m0{width:100%;overflow:hidden;display:-webkit-box;-webkit-line-clamp:3;-webkit-box-orient:vertical;background:#fff}.export-success .bx--inline-notification__text-wrapper{width:100%}h2.svelte-196x1u0.svelte-196x1u0{margin-bottom:.5rem;border-bottom-width:1px;--tw-border-opacity:1;border-color:rgb(209 213 219 / var(--tw-border-opacity));padding-bottom:.5rem}h4.svelte-196x1u0.svelte-196x1u0{margin-bottom:.5rem;margin-top:.5rem}.preview.svelte-196x1u0.svelte-196x1u0{height:26rem}table.svelte-196x1u0 th.svelte-196x1u0{padding-right:1.25rem;text-align:left;font-size:1rem;line-height:1.5rem}table.svelte-196x1u0 td.svelte-196x1u0{padding-right:.75rem;vertical-align:middle}.schema-container.svelte-1xlsh2v{transition:width .2s ease-in-out}.dataset-name.svelte-1xlsh2v{overflow:hidden;white-space:nowrap;text-overflow:ellipsis;max-width:8rem}
