.embedding-badge-nolabel .bx--tooltip__label .bx--tooltip__trigger{margin:0}.embedding-badge-nolabel .bx--tag__custom-icon{margin-right:0}.named-value-name.svelte-1689hje{max-width:15rem}.text-preview-overlay.svelte-1e6kot5{-webkit-mask-image:linear-gradient(to top,transparent,white 75px);mask-image:linear-gradient(to top,transparent,white 75px)}.highlight-span.svelte-1e6kot5{padding-top:1.5px;padding-bottom:1.5px}
