import{S as p,i as A,s as S,D as h,Z as o,_,m,h as d,n as B,$ as g,b as M,a0 as z,K as C,a1 as c,M as v,Q as b,q as E,r as R,u as Z}from"./index.ad1b71d6.js";function w(l){let e,s;return{c(){e=o("title"),s=E(l[1])},l(t){e=_(t,"title",{});var r=m(e);s=R(r,l[1]),r.forEach(d)},m(t,r){M(t,e,r),z(e,s)},p(t,r){r&2&&Z(s,t[1])},d(t){t&&d(e)}}}function k(l){let e,s,t=l[1]&&w(l),r=[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},{width:l[0]},{height:l[0]},l[2],l[3]],n={};for(let i=0;i<r.length;i+=1)n=h(n,r[i]);return{c(){e=o("svg"),t&&t.c(),s=o("path"),this.h()},l(i){e=_(i,"svg",{xmlns:!0,viewBox:!0,fill:!0,preserveAspectRatio:!0,width:!0,height:!0});var a=m(e);t&&t.l(a),s=_(a,"path",{d:!0}),m(s).forEach(d),a.forEach(d),this.h()},h(){B(s,"d","M29,27.5859l-7.5521-7.5521a11.0177,11.0177,0,1,0-1.4141,1.4141L27.5859,29ZM4,13a9,9,0,1,1,9,9A9.01,9.01,0,0,1,4,13Z"),g(e,n)},m(i,a){M(i,e,a),t&&t.m(e,null),z(e,s)},p(i,[a]){i[1]?t?t.p(i,a):(t=w(i),t.c(),t.m(e,s)):t&&(t.d(1),t=null),g(e,n=C(r,[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},a&1&&{width:i[0]},a&1&&{height:i[0]},a&4&&i[2],a&8&&i[3]]))},i:c,o:c,d(i){i&&d(e),t&&t.d()}}}function q(l,e,s){let t,r;const n=["size","title"];let i=v(e,n),{size:a=16}=e,{title:f=void 0}=e;return l.$$set=u=>{s(5,e=h(h({},e),b(u))),s(3,i=v(e,n)),"size"in u&&s(0,a=u.size),"title"in u&&s(1,f=u.title)},l.$$.update=()=>{s(4,t=e["aria-label"]||e["aria-labelledby"]||f),s(2,r={"aria-hidden":t?void 0:!0,role:t?"img":void 0,focusable:Number(e.tabindex)===0?!0:void 0})},e=b(e),[a,f,r,i,t]}class Y extends p{constructor(e){super(),A(this,e,q,k,S,{size:0,title:1})}}export{Y as S};
