import{a1 as l,s as q,ab as k,L as w,a8 as z}from"./index.ad1b71d6.js";const c=[];function A(e,o){return{subscribe:T(e,o).subscribe}}function T(e,o=l){let r;const n=new Set;function u(t){if(q(e,t)&&(e=t,r)){const i=!c.length;for(const s of n)s[1](),c.push(s,e);if(i){for(let s=0;s<c.length;s+=2)c[s][0](c[s+1]);c.length=0}}}function f(t){u(t(e))}function b(t,i=l){const s=[t,i];return n.add(s),n.size===1&&(r=o(u)||l),t(e),()=>{n.delete(s),n.size===0&&r&&(r(),r=null)}}return{set:u,update:f,subscribe:b}}function S(e,o,r){const n=!Array.isArray(e),u=n?[e]:e,f=o.length<2;return A(r,b=>{let t=!1;const i=[];let s=0,p=l;const d=()=>{if(s)return;p();const a=o(n?i[0]:i,b);f?b(a):p=z(a)?a:l},y=u.map((a,_)=>k(a,m=>{i[_]=m,s&=~(1<<_),t&&d()},()=>{s|=1<<_}));return t=!0,d(),function(){w(y),p(),t=!1}})}var g;const x=((g=globalThis.__sveltekit_1nl1p6g)==null?void 0:g.base)??"";var h;const j=((h=globalThis.__sveltekit_1nl1p6g)==null?void 0:h.assets)??x;export{j as a,x as b,S as d,A as r,T as w};
