.concept-score-pill .bx--tooltip__label{margin-right:.25rem;display:inline-block;height:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap;max-width:5rem}.concept-score-pill .bx--tooltip__content{display:flex;flex-direction:column;align-items:center}table.svelte-zc5pc5 td.svelte-zc5pc5{padding:.25rem .5rem}.dataset-link.bx--btn{min-height:0px}
