import{S as V,i as p,s as C,D as m,Z as f,_ as o,m as _,h as u,n as c,$ as v,b as H,a0 as g,K as E,a1 as b,M as w,Q as z,q as A,r as B,u as R}from"./index.ad1b71d6.js";function M(r){let e,a;return{c(){e=f("title"),a=A(r[1])},l(s){e=o(s,"title",{});var t=_(e);a=B(t,r[1]),t.forEach(u)},m(s,t){H(s,e,t),g(e,a)},p(s,t){t&2&&R(a,s[1])},d(s){s&&u(e)}}}function k(r){let e,a,s,t=r[1]&&M(r),h=[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},{width:r[0]},{height:r[0]},r[2],r[3]],n={};for(let i=0;i<h.length;i+=1)n=m(n,h[i]);return{c(){e=f("svg"),t&&t.c(),a=f("path"),s=f("path"),this.h()},l(i){e=o(i,"svg",{xmlns:!0,viewBox:!0,fill:!0,preserveAspectRatio:!0,width:!0,height:!0});var l=_(e);t&&t.l(l),a=o(l,"path",{d:!0}),_(a).forEach(u),s=o(l,"path",{d:!0}),_(s).forEach(u),l.forEach(u),this.h()},h(){c(a,"d","M12 12H14V24H12zM18 12H20V24H18z"),c(s,"d","M4 6V8H6V28a2 2 0 002 2H24a2 2 0 002-2V8h2V6zM8 28V8H24V28zM12 2H20V4H12z"),v(e,n)},m(i,l){H(i,e,l),t&&t.m(e,null),g(e,a),g(e,s)},p(i,[l]){i[1]?t?t.p(i,l):(t=M(i),t.c(),t.m(e,a)):t&&(t.d(1),t=null),v(e,n=E(h,[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},l&1&&{width:i[0]},l&1&&{height:i[0]},l&4&&i[2],l&8&&i[3]]))},i:b,o:b,d(i){i&&u(e),t&&t.d()}}}function q(r,e,a){let s,t;const h=["size","title"];let n=w(e,h),{size:i=16}=e,{title:l=void 0}=e;return r.$$set=d=>{a(5,e=m(m({},e),z(d))),a(3,n=w(e,h)),"size"in d&&a(0,i=d.size),"title"in d&&a(1,l=d.title)},r.$$.update=()=>{a(4,s=e["aria-label"]||e["aria-labelledby"]||l),a(2,t={"aria-hidden":s?void 0:!0,role:s?"img":void 0,focusable:Number(e.tabindex)===0?!0:void 0})},e=z(e),[i,l,t,n,s]}class S extends V{constructor(e){super(),p(this,e,q,k,C,{size:0,title:1})}}export{S as T};
