.bx--inline-loading{width:unset!important;min-height:unset!important}.editor-container.svelte-15hppq8{width:100%}.keyword-search-bg{--tw-bg-opacity:1;background-color:rgb(120 53 15 / var(--tw-bg-opacity));opacity:.2}.keyword-search-glyph{border-right-width:8px;--tw-border-opacity:1;border-color:rgb(120 53 15 / var(--tw-border-opacity));opacity:.2}.keyword-search-text{font-weight:800;--tw-text-opacity:1;color:rgb(139 92 246 / var(--tw-text-opacity));text-decoration-line:underline}.concept-search-bg,.semantic-search-bg{--tw-bg-opacity:.2}.concept-search-text,.semantic-search-text{--tw-text-opacity:1;color:rgb(0 0 0 / var(--tw-text-opacity))}.leaf-bg{--tw-bg-opacity:1;background-color:rgb(194 65 12 / var(--tw-bg-opacity));opacity:.2}.leaf-glyph{border-right-width:8px;--tw-border-opacity:1;border-color:rgb(194 65 12 / var(--tw-border-opacity));opacity:.2}.leaf-text{--tw-text-opacity:1;color:rgb(139 92 246 / var(--tw-text-opacity))}.line-selection-bg,.line-col-selection-bg{background-color:rgb(107 114 128 / var(--tw-bg-opacity));--tw-bg-opacity:.2}.line-selection-glyph{border-right-width:8px;--tw-border-opacity:1;border-color:rgb(107 114 128 / var(--tw-border-opacity));opacity:.5}.editor-container .monaco-editor .lines-content.monaco-editor-background{margin-left:10px}.active.svelte-1tlnk00{--tw-bg-opacity:1;background-color:rgb(233 213 255 / var(--tw-bg-opacity));outline-color:#7e22ce}
